part of enhanced_ddp;

enum ConnectStatus {
  disconnected,
  dialing,
  connecting,
  connected,
}

// ignore: prefer_generic_function_type_aliases
typedef void _MessageHandler(Map<String, dynamic> message);

typedef ConnectionListener = void Function();
typedef StatusListener = void Function(ConnectStatus status);

abstract class ConnectionNotifier {
  void addConnectionListener(ConnectionListener listener);
}

abstract class StatusNotifier {
  void addStatusListener(StatusListener listener);
  void removeStatusListener(StatusListener listener);
}

abstract class DdpReconnectListener {
  void onReconnectBegin();
  void onReconnectDone();
  void onConnected();
}

class ReconnectListenersHolder implements DdpReconnectListener {
  // ignore: prefer_final_fields
  List<DdpReconnectListener> _listeners = [];

  void addReconnectListener(DdpReconnectListener listener) {
    if (listener == null) {
      return;
    }

    removeReconnectListener(listener);
    _listeners.add(listener);
  }

  void removeReconnectListener(DdpReconnectListener listener) {
    if (listener == null) {
      return;
    }
    _listeners.remove(listener);
  }

// ignore: annotate_overrides
  void onReconnectBegin() {
    _listeners.forEach((listener) {
      try {
        listener.onReconnectBegin();
      } catch (exception) {
        print(exception);
      }
    });
  }

// ignore: annotate_overrides
  void onReconnectDone() {
    _listeners.forEach((listener) {
      try {
        listener.onReconnectDone();
      } catch (exception) {
        print(exception);
      }
    });
  }

  // ignore: annotate_overrides
  void onConnected() {
    _listeners.forEach((listener) {
      try {
        listener.onConnected();
      } catch (exception) {
        print(exception);
      }
    });
  }
}

class DdpClient implements ConnectionNotifier, StatusNotifier {
  // ignore: prefer_final_fields
  String _name;
  Duration heartbeatInterval;
  Duration heartbeatTimeout;
  Duration reconnectInterval;

  WriterStats _writeSocketStats;
  WriterStats _writeStats;
  WriterLogger _writeLog;
  ReaderStats _readSocketStats;
  ReaderStats _readStats;
  ReaderLogger _readLog;

  int _reconnects;
  int _pingsIn;
  int _pingsOut;

  String _session;
  String _version;
  // ignore: unused_field
  String _serverId;
  WebSocketChannel _ws;
  String _url;
  // ignore: unused_field
  String _origin;

  Timer _pingTimer;

  Map<String, List<_PingTracker>> _pings;
  Map<String, Call> _calls;
  Map<String, Call> _subs;
  Map<String, Call> _unsubs;

  Map<String, Collection> _collections;
  ConnectStatus _connectionStatus;
  Timer _reconnectTimer;

  List<StatusListener> _statusListeners;
  List<ConnectionListener> _connectionListener;

  _IdManager _idManager;

  Map<String, _MessageHandler> _messageHandlers;

  // ignore: prefer_final_fields
  ReconnectListenersHolder _reconnectListenersHolder =
      ReconnectListenersHolder();

  bool _waitingForConnect = false;

  DdpClient(_name, String url, String origin) {
    heartbeatInterval = const Duration(seconds: 15);
    heartbeatTimeout = const Duration(seconds: 20);
    reconnectInterval = const Duration(seconds: 5);

    _collections = {};
    _url = url;
    _pings = {};
    _calls = {};
    _subs = {};
    _unsubs = {};
    _connectionStatus = ConnectStatus.disconnected;

    _writeSocketStats = WriterStats(null);
    _writeStats = WriterStats(null);
    _readSocketStats = ReaderStats(null);
    _readStats = ReaderStats(null);

    _writeLog = WriterLogger.text(null);
    _readLog = ReaderLogger.text(null);

    _idManager = _IdManager();
    _statusListeners = [];
    _connectionListener = [];

    _reconnects = 0;
    _pingsIn = 0;
    _pingsOut = 0;
  }
  void addReconnectListener(DdpReconnectListener listener) {
    _reconnectListenersHolder.addReconnectListener(listener);
  }

  void removeReconnectListener(DdpReconnectListener listener) {
    _reconnectListenersHolder.removeReconnectListener(listener);
  }

  void _log(String msg) {
    print('[DdpClient - ${_name}] $msg');
  }

  String get session => _session;

  String get version => _version;

  @override
  void addConnectionListener(ConnectionListener listener) {
    _connectionListener.add(listener);
  }

  @override
  void addStatusListener(StatusListener listener) {
    _statusListeners.add(listener);
  }

  @override
  void removeStatusListener(StatusListener listener) {
    _statusListeners.remove(listener);
  }

  void _status(ConnectStatus status) {
    if (_connectionStatus == status) {
      return;
    }
    _connectionStatus = status;
    _statusListeners.forEach((l) => l(status));
  }

  void connect() async {
    try {
      if (_waitingForConnect == false) {
        _waitingForConnect = true;
        _reconnectListenersHolder.onReconnectBegin();
      }
      _status(ConnectStatus.dialing);
      final ws = await WebSocketChannel.connect(Uri.parse(_url));
      _start(ws, Message.connect());
      _waitingForConnect = false;
      _reconnectListenersHolder.onConnected();
    } catch (error) {
      print('DDP ERROR (on connect): $error');
      _reconnectLater();
    }
  }

  void reconnect() async {
    try {
      if (_reconnectTimer != null) {
        _reconnectTimer.cancel();
        _reconnectTimer = null;
      }

      if (_waitingForConnect == false) {
        _waitingForConnect = true;
        _reconnectListenersHolder.onReconnectBegin();
      }

      close();
      _reconnects++;
      _status(ConnectStatus.dialing);
      final connection = await WebSocketChannel.connect(Uri.parse(_url));
      _start(connection, Message.reconnect(_session));
      _calls.values.forEach((call) => send(
          Message.method(call.id, call.serviceMethod, call.args).toJson()));
      _subs.values.forEach((call) =>
          send(Message.sub(call.id, call.serviceMethod, call.args).toJson()));
      _waitingForConnect = false;
      _reconnectListenersHolder.onConnected();
    } catch (error) {
      print('DDP ERROR (on reconnect): $error');
      close();
      _reconnectLater();
    }
  }

  void onAuthDone() {
    _reconnectListenersHolder.onReconnectDone();
  }

  Call subscribe(String subName, OnCallDone done, List<dynamic> args) {
    args ??= [];

    final call = Call()
      ..id = subName + '-' + _idManager.next()
      ..serviceMethod = subName
      ..args = args
      ..owner = this;

    done ??= (c) {};

    call.onceDone(done);

    _subs[call.id] = call;

    send(Message.sub(call.id, subName, args).toJson());
    return call;
  }

  Future<Call> sub(String subName, List<dynamic> args) {
    final completer = Completer<Call>();
    subscribe(subName, (call) => completer.complete(call), args);
    return completer.future;
  }

  Call unSubscribe(String id, OnCallDone done) {
    final call = Call()
      ..id = id
      ..owner = this;

    done ??= (c) {};
    call.onceDone(done);
    _unsubs[call.id] = call;
    send(Message.unSub(call.id).toJson());
    return call;
  }

  Future<Call> unSub(String id) {
    final completer = Completer<Call>();
    unSubscribe(id, (call) => completer.complete(call));
    return completer.future;
  }

  Call go(String serviceMethod, OnCallDone done, List<dynamic> args) {
    args ??= [];
    final call = Call()
      ..id = _idManager.next()
      ..serviceMethod = serviceMethod
      ..args = args
      ..owner = this;
    done ??= (c) {};
    call.onceDone(done);
    _calls[call.id] = call;
    send(Message.method(call.id, serviceMethod, args).toJson());
    return call;
  }

  Future<Call> call(String serviceMethod, List<dynamic> args) {
    final completer = Completer<Call>();
    go(serviceMethod, (call) => completer.complete(call), args);
    return completer.future;
  }

  void send(dynamic msg) {
    _writeStats.add(msg);
  }

  void close() {
    if (_pingTimer != null) {
      _pingTimer.cancel();
      _pingTimer = null;
    }

    if (_ws != null) {
      _ws.closeCode;
      _ws = null;
    }

    _collections.values.forEach((collection) => collection._reset());
    _status(ConnectStatus.disconnected);
  }

  void resetStats() {
    _readSocketStats.reset();
    _readStats.reset();
    _writeSocketStats.reset();
    _writeStats.reset();
    _reconnects = 0;
    _pingsIn = 0;
    _pingsOut = 0;
  }

  ClientStats stats() {
    return ClientStats()
      ..reads = _readSocketStats.snapshot()
      ..totalReads = _readStats.snapshot()
      ..writes = _writeSocketStats.snapshot()
      ..totalWrites = _writeStats.snapshot()
      ..reconnects = _reconnects
      ..pingsSent = _pingsOut
      ..pingsRecv = _pingsIn;
  }

  bool socketLogActive() {
    return _writeLog.active;
  }

  void setSocketLogActive(bool active) {
    assert(active != null);
    _writeLog.active = active;
    _readLog.active = active;
  }

  Collection collectionByName(String name) {
    if (!_collections.containsKey(name)) {
      final collection = Collection.key(name);
      _collections[name] = collection;
    }
    return _collections[name];
  }

  List<CollectionStats> collectionStats() {
    List<CollectionStats> stats = [];
    _collections.forEach((name, collection) => stats.add(CollectionStats()
      ..name = name
      ..count = collection.findAll().length));
    return stats;
  }

  void _start(WebSocketChannel ws, _Connect connect) {
    _status(ConnectStatus.connecting);

    _initMessageHandlers();
    _ws = ws;
    _writeLog.setWriter(ws.sink);
    _writeSocketStats = WriterStats(_writeLog);
    _writeStats.setWriter(ws.sink);
    _readLog.setReader(ws.stream);
    _readSocketStats = ReaderStats(_readLog);
    _readStats.setReader(_readSocketStats);

    inboxManager();

    send(connect.toJson());
  }

  void _reconnectLater() {
    close();
    _reconnectTimer ??= Timer(reconnectInterval, reconnect);
  }

  void ping() {
    pingPong(_idManager.next(), heartbeatTimeout, (err) {
      if (err != null) {
        _reconnectLater();
      }
    });
  }

  void pingPong(String id, Duration timeout, Function(Error) handler) {
    send(Message.ping(id).toJson());
    _pingsOut++;
    if (!_pings.containsKey(id)) {
      _pings[id] = [];
    }
    final pingTracker = _PingTracker()
      .._handler = handler
      .._timeout = timeout
      .._timer = Timer(timeout, () {
        handler(ArgumentError('ping timeout'));
      });
    _pings[id].add(pingTracker);
  }

  void _initMessageHandlers() {
    _messageHandlers = {};
    _messageHandlers['connected'] = (msg) {
      _status(ConnectStatus.connected);
      _collections.values.forEach((c) => c._init());
      _version = '1';
      _session = msg['session'] as String;
      _pingTimer = Timer.periodic(heartbeatInterval, (Timer timer) {
        ping();
      });
      _connectionListener.forEach((l) => l());
    };
    _messageHandlers['ping'] = (msg) {
      if (msg.containsKey('id')) {
        send(Message.pong(msg['id']).toJson());
      } else {
        send(Message.pong(null).toJson());
      }
      _pingsIn++;
    };
    _messageHandlers['pong'] = (msg) {
      var key = '';
      if (msg.containsKey('id')) {
        key = msg['id'] as String;
      }
      if (_pings.containsKey(key)) {
        final pings = _pings[key];
        if (pings.isNotEmpty) {
          final ping = pings[0];
          final newPings = pings.sublist(1);
          if (key.isEmpty || pings.isNotEmpty) {
            _pings[key] = newPings;
          }
          ping._timer.cancel();
          ping._handler(null);
        }
      }
    };
    _messageHandlers['nosub'] = (msg) {
      if (msg.containsKey('id')) {
        final id = msg['id'] as String;
        final runningSub = _subs[id];
        if (runningSub != null) {
          print(runningSub);
          _log('Subscription returned a nosub error $msg');
          runningSub.error =
              ArgumentError('Subscription returned a nosub error');
          runningSub.done();
          _subs.remove(id);
        }

        final runningUnSub = _unsubs[id];
        if (runningUnSub != null) {
          runningUnSub.done();
          _unsubs.remove(id);
        }
      }
    };
    _messageHandlers['ready'] = (msg) {
      if (msg.containsKey('subs')) {
        final subs = msg['subs'] as List<dynamic>;
        subs.forEach((sub) {
          if (_subs.containsKey(sub)) {
            _subs[sub].done();
            _subs.remove(sub);
          }
        });
      }
    };
    _messageHandlers['added'] = (msg) => _collectionBy(msg)._added(msg);
    _messageHandlers['changed'] = (msg) => _collectionBy(msg)._changed(msg);
    _messageHandlers['removed'] = (msg) => _collectionBy(msg)._removed(msg);
    _messageHandlers['addedBefore'] =
        (msg) => _collectionBy(msg)._addedBefore(msg);
    _messageHandlers['movedBefore'] =
        (msg) => _collectionBy(msg)._movedBefore(msg);
    _messageHandlers['result'] = (msg) {
      if (msg.containsKey('id')) {
        final id = msg['id'];
        final call = _calls[id];
        _calls.remove(id);
        if (msg.containsKey('error')) {
          if (msg['error'] != null) {
            final e = msg['error'];
            call.error = ArgumentError(json.encode(e));
            call.reply = e;
          }
        } else {
          call.reply = msg['result'];
        }
        call.done();
      }
    };
    _messageHandlers['updated'] = (msg) {};
  }

  void inboxManager() {
    _readStats.listen((event) {
      final message = json.decode(event) as Map<String, dynamic>;
      if (message.containsKey('msg')) {
        final mtype = message['msg'];
        if (_messageHandlers.containsKey(mtype)) {
          _messageHandlers[mtype](message);
        } else {
          _log('Server sent unexpected message ${message}');
        }
      } else if (message.containsKey('server_id')) {
        final serverId = message['server_id'];
        if (serverId.runtimeType == String) {
        } else {
          _log('Server cluster node ${serverId}');
        }
      } else {
        _log('Server sent message without `msg` field ${message}');
      }
    }, onDone: _onDone, onError: _onError, cancelOnError: true);
  }

  void _onDone() {
    _status(ConnectStatus.disconnected);
    print('Disconnect due to websocket onDone');
    print('Disconnected code: ${_ws.closeCode}, reason: ${_ws.closeReason}');
    print('Schedule reconnect due to websocket onDone');
    _reconnectLater();
  }

  void _onError(dynamic error) {
    _status(ConnectStatus.disconnected);
    print('Disconnect due to websocket onError');
    print('Error: $error');
    print('Schedule reconnect due to websocket onError');
    _reconnectLater();
  }

  Collection _collectionBy(Map<String, dynamic> msg) {
    if (msg.containsKey('collection')) {
      final name = msg['collection'];
      if (name.runtimeType == String) {
        return collectionByName(name);
      }
    }
    return Collection.mock();
  }
}
