part of enhanced_ddp;

enum LoggingDataType {
  DataByte,
  DataText,
}

class ReaderProxy extends Stream<dynamic> {
  Stream<dynamic> _reader;

  ReaderProxy(_reader);

  @override
  StreamSubscription listen(
    void Function(dynamic event) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return _reader.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  void setReader(Stream<dynamic> reader) {
    _reader = reader;
  }
}

class WriterProxy implements StreamSink<dynamic> {
  StreamSink<dynamic> _writer;

  WriterProxy(_writer);

  @override
  void add(event) {
    _writer.add(event);
  }

  @override
  void addError(Object error, [StackTrace stackTrace]) {
    _writer.addError(error, stackTrace);
  }

  @override
  Future addStream(Stream stream) {
    return _writer.addStream(stream);
  }

  @override
  Future close() {
    return _writer.close();
  }

  @override
  Future get done {
    return _writer.done;
  }

  void setWriter(WebSocketSink writer) {
    _writer = writer;
  }
}

typedef _Logger = void Function(Object object);

class LoggerMixin {
  bool active;
  int truncate;
  LoggingDataType _dtype;
  _Logger _logger;

  int log(dynamic p, int n) {
    if (this.active) {
      int limit = n;
      bool trancated = false;
      if (truncate > 0 && truncate < limit) {
        limit = truncate;
        trancated = true;
      }
      switch (_dtype) {
        case LoggingDataType.DataText:
          if (trancated) {
            if (p is List) {
              _logger('[${n}] ${utf8.decode(p.sublist(0, limit))}...');
            } else if (p.runtimeType == String) {
              _logger('[${n}] ${p.substring(0, limit)}...');
            }
          } else {
            if (p is List) {
              _logger('[${n}] ${utf8.decode(p.sublist(0, limit))}');
            } else if (p.runtimeType == String) {
              _logger('[${n}] ${p.substring(0, limit)}');
            }
          }
          break;
        case LoggingDataType.DataByte:
        default:
        // Don't know what to do? maybe dataByte is not necessary to log?
      }
    }
    return n;
  }
}

class ReaderLogger extends ReaderProxy with LoggerMixin {
  ReaderLogger(Stream reader) : super(reader);

  factory ReaderLogger.text(Stream reader) => ReaderLogger(reader)
    .._logger = ((Object obj) => print('<- $obj'))
    ..active = true
    .._dtype = LoggingDataType.DataText
    ..truncate = 80;

  @override
  StreamSubscription listen(
    void Function(dynamic event) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return super.listen(
      (event) {
        log(event, event.length);
        onData(event);
      },
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }
}

class WriterLogger extends WriterProxy with LoggerMixin {
  WriterLogger(StreamSink writer) : super(writer);

  factory WriterLogger.text(StreamSink writer) => WriterLogger(writer)
    .._logger = ((Object obj) => print('-> $obj'))
    ..active = true
    .._dtype = LoggingDataType.DataText
    ..truncate = 80;

  @override
  void add(event) {
    log(event, event.length);
    super.add(event);
  }
}

class Stats {
  int bytes;
  int ops;
  int errors;
  Duration runtime;
}

class ClientStats {
  Stats reads;
  Stats totalReads;
  Stats writes;
  Stats totalWrites;
  int reconnects;
  int pingsSent;
  int pingsRecv;
}

class CollectionStats {
  String name;
  int count;
}

class StatsTrackerMixin {
  int _bytes;
  int _ops;
  int _errors;
  DateTime _start;

  int op(int n) {
    _ops++;
    _bytes += n;

    return n;
  }

  Stats snapshot() {
    return _snap();
  }

  Stats reset() {
    final stats = _snap();
    _bytes = 0;
    _ops = 0;
    _errors = 0;
    _start = DateTime.now();
    return stats;
  }

  Stats _snap() {
    return Stats()
      ..bytes = _bytes
      ..ops = _ops
      ..errors = _errors
      ..runtime = DateTime.now().difference(_start);
  }
}

class ReaderStats extends ReaderProxy with StatsTrackerMixin {
  ReaderStats(Stream reader) : super(reader) {
    _bytes = 0;
    _ops = 0;
    _errors = 0;
    _start = DateTime.now();
  }

  @override
  StreamSubscription listen(
    void Function(dynamic event) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return super.listen(
      (event) {
        op(event.length);
        onData(event);
      },
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }
}

class WriterStats extends WriterProxy with StatsTrackerMixin {
  WriterStats(StreamSink writer) : super(writer) {
    _bytes = 0;
    _ops = 0;
    _errors = 0;
    _start = DateTime.now();
  }

  @override
  void add(dynamic event) {
    op(event.length);
    super.add(event);
  }
}
